@@ -123,6 +123,9 @@ class Protocol(threading.local):
             self.host = self.port = None
             self.set_retry_delay(0)
 
+    def __str__(self):
+        return "{}_{}_{}".format(self.server, self._username, self._password)
+
     @property
     def server_uses_unix_socket(self):
         return self.host is None
@@ -423,11 +426,12 @@ class Protocol(threading.local):
         :rtype: object
         """
         logger.debug('Getting key %s', key)
+        keybytes = str_to_bytes(key)
         data = struct.pack(self.HEADER_STRUCT +
-                           self.COMMANDS['get']['struct'] % (len(key)),
+                           self.COMMANDS['get']['struct'] % (len(keybytes),),
                            self.MAGIC['request'],
                            self.COMMANDS['get']['command'],
-                           len(key), 0, 0, 0, len(key), 0, 0, str_to_bytes(key))
+                           len(keybytes), 0, 0, 0, len(keybytes), 0, 0, keybytes)
         self._send(data)
 
         (magic, opcode, keylen, extlen, datatype, status, bodylen, opaque,
@@ -487,30 +491,25 @@ class Protocol(threading.local):
         it only decoded those keys back to string which were sent as string
 
         :param keys: A list of keys to from server.
-        :type keys: list
+        :type keys: Collection
         :return: A dict with all requested keys.
         :rtype: dict
         """
         # pipeline N-1 getkq requests, followed by a regular getk to uncork the
         # server
-        o_keys = keys
-        keys, last = keys[:-1], str_to_bytes(keys[-1])
-        if six.PY2:
-            msg = ''
-        else:
-            msg = b''
-        msg = msg.join([
-            struct.pack(self.HEADER_STRUCT +
-                        self.COMMANDS['getkq']['struct'] % (len(key)),
-                        self.MAGIC['request'],
-                        self.COMMANDS['getkq']['command'],
-                        len(key), 0, 0, 0, len(key), 0, 0, str_to_bytes(key))
-            for key in keys])
-        msg += struct.pack(self.HEADER_STRUCT +
-                           self.COMMANDS['getk']['struct'] % (len(last)),
-                           self.MAGIC['request'],
-                           self.COMMANDS['getk']['command'],
-                           len(last), 0, 0, 0, len(last), 0, 0, last)
+        n = len(keys)
+        if n == 0:
+            return {}
+
+        msg = b''
+        for i, key in enumerate(keys):
+            keybytes = str_to_bytes(key)
+            command = self.COMMANDS['getk' if i == n - 1 else 'getkq']
+            msg += struct.pack(self.HEADER_STRUCT +
+                               command['struct'] % (len(keybytes),),
+                               self.MAGIC['request'],
+                               command['command'],
+                               len(keybytes), 0, 0, 0, len(keybytes), 0, 0, keybytes)
 
         self._send(msg)
 
@@ -524,25 +523,19 @@ class Protocol(threading.local):
                 flags, key, value = struct.unpack('!L%ds%ds' %
                                                   (keylen, bodylen - keylen - 4),
                                                   extra_content)
-                if six.PY2:
-                    d[key] = self.deserialize(value, flags), cas
-                else:
-                    try:
-                        decoded_key = key.decode()
-                    except UnicodeDecodeError:
-                        d[key] = self.deserialize(value, flags), cas
-                    else:
-                        if decoded_key in o_keys:
-                            d[decoded_key] = self.deserialize(value, flags), cas
-                        else:
-                            d[key] = self.deserialize(value, flags), cas
+                d[key] = self.deserialize(value, flags), cas
 
             elif status == self.STATUS['server_disconnected']:
                 break
             elif status != self.STATUS['key_not_found']:
                 raise MemcachedException('Code: %d Message: %s' % (status, extra_content), status)
 
-        return d
+        ret = {}
+        for key in keys:
+            keybytes = str_to_bytes(key)
+            if keybytes in d:
+                ret[key] = d[keybytes]
+        return ret
 
     def _set_add_replace(self, command, key, value, time, cas=0, compress_level=-1):
         """
@@ -570,12 +563,13 @@ class Protocol(threading.local):
         if isinstance(value, text_type):
             value = value.encode('utf8')
 
+        keybytes = str_to_bytes(key)
         self._send(struct.pack(self.HEADER_STRUCT +
-                               self.COMMANDS[command]['struct'] % (len(key), len(value)),
+                               self.COMMANDS[command]['struct'] % (len(keybytes), len(value)),
                                self.MAGIC['request'],
                                self.COMMANDS[command]['command'],
-                               len(key), 8, 0, 0, len(key) + len(value) + 8, 0, cas, flags,
-                               time, str_to_bytes(key), value))
+                               len(keybytes), 8, 0, 0, len(keybytes) + len(value) + 8, 0, cas, flags,
+                               time, keybytes, value))
 
         (magic, opcode, keylen, extlen, datatype, status, bodylen, opaque,
          cas, extra_content) = self._get_response()
@@ -692,13 +686,13 @@ class Protocol(threading.local):
             0 = no compression, 1 = fastest, 9 = slowest but best,
             -1 = default compression level.
         :type compress_level: int
-        :return: True
-        :rtype: bool
+        :return: List of keys that failed to be set.
+        :rtype: list
         """
-        mappings = mappings.items()
+        mappings = list(mappings.items())
         msg = []
 
-        for key, value in mappings:
+        for opaque, (key, value) in enumerate(mappings):
             if isinstance(key, tuple):
                 key, cas = key
             else:
@@ -711,14 +705,15 @@ class Protocol(threading.local):
             else:
                 command = 'setq'
 
+            keybytes = str_to_bytes(key)
             flags, value = self.serialize(value, compress_level=compress_level)
             m = struct.pack(self.HEADER_STRUCT +
-                            self.COMMANDS[command]['struct'] % (len(key), len(value)),
+                            self.COMMANDS[command]['struct'] % (len(keybytes), len(value)),
                             self.MAGIC['request'],
                             self.COMMANDS[command]['command'],
-                            len(key),
-                            8, 0, 0, len(key) + len(value) + 8, 0, cas or 0,
-                            flags, time, str_to_bytes(key), value)
+                            len(keybytes),
+                            8, 0, 0, len(keybytes) + len(value) + 8, opaque, cas or 0,
+                            flags, time, keybytes, value)
             msg.append(m)
 
         m = struct.pack(self.HEADER_STRUCT +
@@ -728,24 +723,26 @@ class Protocol(threading.local):
                         0, 0, 0, 0, 0, 0, 0)
         msg.append(m)
 
-        if six.PY2:
-            msg = ''.join(msg)
-        else:
-            msg = b''.join(msg)
+        msg = b''.join(msg)
 
         self._send(msg)
 
         opcode = -1
-        retval = True
+        failed = []
         while opcode != self.COMMANDS['noop']['command']:
             (magic, opcode, keylen, extlen, datatype, status, bodylen, opaque,
              cas, extra_content) = self._get_response()
-            if status != self.STATUS['success']:
-                retval = False
             if status == self.STATUS['server_disconnected']:
-                break
+                # Assume that the entire operation failed.
+                return list(key for key, value in mappings)
+            if status != self.STATUS['success']:
+                key, value = mappings[opaque]
+                if isinstance(key, tuple):
+                    failed.append((key[0], cas))
+                else:
+                    failed.append(key)
 
-        return retval
+        return failed
 
     def _incr_decr(self, command, key, value, default, time):
         """
@@ -762,14 +759,15 @@ class Protocol(threading.local):
         :return: Actual value of the key on server
         :rtype: int
         """
+        keybytes = str_to_bytes(key)
         time = time if time >= 0 else self.MAXIMUM_EXPIRE_TIME
         self._send(struct.pack(self.HEADER_STRUCT +
                                self.COMMANDS[command]['struct'] % len(key),
                                self.MAGIC['request'],
                                self.COMMANDS[command]['command'],
-                               len(key),
-                               20, 0, 0, len(key) + 20, 0, 0, value,
-                               default, time, str_to_bytes(key)))
+                               len(keybytes),
+                               20, 0, 0, len(keybytes) + 20, 0, 0, value,
+                               default, time, keybytes))
 
         (magic, opcode, keylen, extlen, datatype, status, bodylen, opaque,
          cas, extra_content) = self._get_response()
@@ -828,11 +826,12 @@ class Protocol(threading.local):
         :rtype: bool
         """
         logger.debug('Deleting key %s', key)
+        keybytes = str_to_bytes(key)
         self._send(struct.pack(self.HEADER_STRUCT +
-                               self.COMMANDS['delete']['struct'] % len(key),
+                               self.COMMANDS['delete']['struct'] % (len(keybytes),),
                                self.MAGIC['request'],
                                self.COMMANDS['delete']['command'],
-                               len(key), 0, 0, 0, len(key), 0, cas, str_to_bytes(key)))
+                               len(keybytes), 0, 0, 0, len(keybytes), 0, cas, keybytes))
 
         (magic, opcode, keylen, extlen, datatype, status, bodylen, opaque,
          cas, extra_content) = self._get_response()
@@ -855,17 +854,15 @@ class Protocol(threading.local):
         :rtype: bool
         """
         logger.debug('Deleting keys %r', keys)
-        if six.PY2:
-            msg = ''
-        else:
-            msg = b''
+        msg = b''
         for key in keys:
+            keybytes = str_to_bytes(key)
             msg += struct.pack(
                 self.HEADER_STRUCT +
-                self.COMMANDS['delete']['struct'] % len(key),
+                self.COMMANDS['delete']['struct'] % (len(keybytes),),
                 self.MAGIC['request'],
                 self.COMMANDS['delete']['command'],
-                len(key), 0, 0, 0, len(key), 0, 0, str_to_bytes(key))
+                len(keybytes), 0, 0, 0, len(keybytes), 0, 0, keybytes)
 
         msg += struct.pack(
             self.HEADER_STRUCT +
